import shutil
import itertools
import unittest

from dampr import Dampr, BlockMapper, BlockReducer, Dataset
from dampr.inputs import UrlsInput

class DamprTest(unittest.TestCase):

    def setUp(self):
        items = list(range(10, 20))
        self.items = Dampr.memory(items, partitions=2)

    def test_identity(self):
        results = self.items.run()
        self.assertEquals(list(range(10, 20)), list(results))

    def test_count(self):
        res = self.items \
                .group_by(lambda x: 1, lambda x: 1) \
                .reduce(lambda k, it: sum(it)) \
                .run()

        self.assertEquals(10, next(iter(res))[1])

    def test_count_red(self):
        res = self.items \
                .count(lambda x: None) \
                .run()

        self.assertEquals((None, 10), next(iter(res)))

    def test_sum(self):
        res = self.items \
                .group_by(lambda x: 1) \
                .reduce(lambda k, it: sum(it)) \
                .run()

        self.assertEquals(sum(range(10, 20)), next(iter(res))[1])

        # Sum odds/evens
        res = self.items \
                .group_by(lambda v: v % 2) \
                    .reduce(lambda k, it: sum(it)) \
                .run()

        evens_odds = [kv[1] for kv in res]
        self.assertEquals([10+12+14+16+18, 11+13+15+17+19], evens_odds)

    def test_filter(self):
        odds = self.items \
                .filter(lambda i: i % 2 == 1) \
                .run()

        odds = list(odds)
        self.assertEquals([11, 13, 15, 17, 19], odds)

    def test_sort(self):
        res = self.items.sort_by(lambda x: -x).run()
        self.assertEquals([19,18,17,16,15,14,13,12,11,10],list(res))

    def test_reduce_join(self):
        items2 = Dampr.memory(list(range(10)))
        res = self.items \
                .group_by(lambda x: x % 2) \
                .join(items2.group_by(lambda x: x % 2)) \
                    .reduce(lambda l, r: list(sorted(itertools.chain(l, r)))) \
                .run()

        output = list(res)
        self.assertEquals((0, [0,2,4,6,8,10,12,14,16,18]), output[0])
        self.assertEquals((1, [1,3,5,7,9,11,13,15,17,19]), output[1])

    def test_disjoint(self):
        items2 = Dampr.memory(list(range(10))) \
                .group_by(lambda x: -x)
        output = self.items.group_by(lambda x: x) \
                .join(items2) \
                .run()
        output = [v for k,v in output]
        self.assertEquals([], output)

    def test_repartition(self):
        items2 = Dampr.memory(list(range(10))) \
                .group_by(lambda x: -x) \
                    .reduce(lambda k, vs: sum(vs))

        output = self.items.group_by(lambda x: x) \
                .join(items2) \
                .run()

        output = [v for k,v in output]
        self.assertEquals([], output)

    def test_associative_reduce(self):
        output = self.items \
                .a_group_by(lambda x: x % 2) \
                    .reduce(lambda x,y: x + y) \
                .run()

        output = list(output)
        self.assertEquals(10 + 12 + 14 + 16 + 18, output[0][1])
        self.assertEquals(11 + 13 + 15 + 17 + 19, output[1][1])
        
    def test_left_join(self):
        to_remove = Dampr.memory(list(range(10, 13)))
        
        output = self.items.group_by(lambda x: x) \
                .join(to_remove.group_by(lambda x: x)) \
                    .left_reduce(lambda l, r: (list(l), list(r))) \
                .filter(lambda llrs: len(llrs[1][1]) == 0) \
                .map(lambda llrs: llrs[1][0][0]) \
                .sort_by(lambda x: x) \
                .run()

        output = list(output)
        self.assertEquals(list(range(13,20)), output)

    def test_combine(self):
        even = self.items.filter(lambda x: x % 2 == 0)
        odd  = self.items.filter(lambda x: x % 2 == 1)

        even_ve, odd_ve = Dampr.run(even, odd)
        self.assertEquals([10,12,14,16,18], list(even_ve))
        self.assertEquals([11,13,15,17,19], list(odd_ve))

    def test_reduce_many(self):
        even = self.items.filter(lambda x: x % 2 == 0)
        odd  = self.items.filter(lambda x: x % 2 == 1)

        def cross(x, y):
            y = list(y)
            for xi in x:
                for yi in y:
                    yield xi * yi

        results = even.group_by(lambda x: 1) \
            .join(odd.group_by(lambda x: 1)) \
                .reduce(cross, many=True) \
            .run().read()

        results = sorted(results)
        e = [10,12,14,16,18]
        o = [11,13,15,17,19]
        expected = sorted([(1, ei * oi) for ei in e for oi in o])
        self.assertEquals(results, expected)

    def test_fold_by(self):
        output = self.items \
                .fold_by(lambda x: 1, 
                        value=lambda x: x % 2, 
                        binop=lambda x, y: x + y)

        results = list(output.run())
        self.assertEquals([(1, 5)], results)

    def test_empty_map(self):
        """
        Test that empty mrcs writes don't blow up
        """
        output = self.items \
                .sample(0.0) \
                .fold_by(lambda x: 1, 
                        value=lambda x: x % 2, 
                        binop=lambda x, y: x + y)

        results = list(output.run())
        self.assertEquals([], results)

    def test_sink(self):
        """
        Test Sinking to disk works and doesn't delete on cleanup
        """
        path = "/tmp/dampr_test_sink"
        sink = self.items.map(lambda x: str(x)) \
                .sink(path=path)

        output = sink.count()

        results = list(output.run())
        self.assertEquals([('{}'.format(i), 1) for i in range(10, 20)], results)

        shutil.rmtree(path)

    def test_cached(self):
        """
        Tests caching stages works
        """
        sink = self.items.map(lambda x: str(x)) \
                .cached()

        res = sink.run()
        output = sink.count()

        results = list(output.run())
        self.assertEquals([(str(i), 1) for i in range(10, 20)], results)

    def test_cross_join(self):
        """
        Tests cross product
        """
        total = self.items \
                .a_group_by(lambda x: 1) \
                    .sum()

        output = self.items \
                .cross_right(total, lambda v1, v2: round(v1 / float(v2[1]), 4)) \
                .sort_by(lambda x: x)

        results = list(output.run())
        count = sum(range(10, 20))
        expected = [round(i / float(count), 4) for i in range(10, 20)]
        self.assertEquals(expected, results)

    def test_cross_join_multi(self):
        """
        Tests cross product of multiple values
        """
        output = self.items \
                .cross_left(self.items, lambda v1, v2: v1 * v2)

        results = sorted(list(output.run()))
        expected = sorted([i * k for i in range(10, 20) for k in range(10, 20)])
        self.assertEquals(expected, results)

    def test_blocks(self):
        """
        Tests Custom Blocks
        """
        from collections import defaultdict
        import heapq
        class TopKMapper(BlockMapper):
            def __init__(self, k):
                self.k = k

            def start(self):
                self.heap = []

            def add(self, _k, lc):
                heapq.heappush(self.heap, (lc[1], lc[0]))
                if len(self.heap) > self.k:
                    heapq.heappop(self.heap)

                return iter([])

            def finish(self):
                for cl in self.heap:
                    yield 1, cl

        class TopKReducer(BlockReducer):
            def __init__(self, k):
                self.k = k

            def start(self):
                pass

            def add(self, k, it):
                for count, letter in heapq.nlargest(self.k, it):
                    yield letter, (letter, count)

        word = Dampr.memory(["supercalifragilisticexpialidociousa"])
        letter_counts = word.flat_map(lambda w: list(w)).count()

        topk = letter_counts \
                .custom_mapper(TopKMapper(2)) \
                .custom_reducer(TopKReducer(2))

        results = sorted(list(topk.run()))
        self.assertEquals(results, [('a', 4), ('i', 7)])

    def test_stream_blocks(self):
        """
        Tests stream blocks
        """
        import heapq
        def map_topk(it):
            heap = []
            for symbol, count in it:
                heapq.heappush(heap, (count, symbol))
                if len(heap) > 2:
                    heapq.heappop(heap)

            return ((1, x) for x in heap)

        def reduce_topk(it):
            counts = (v for k, vit in it for v in vit)
            for count, symbol in heapq.nlargest(2, counts):
                yield symbol, count

        word = Dampr.memory(["supercalifragilisticexpialidociousa"])
        letter_counts = word.flat_map(lambda w: list(w)).count()

        topk = letter_counts \
                .partition_map(map_topk) \
                .partition_reduce(reduce_topk)

        results = sorted(list(topk.run()))
        self.assertEquals(results, [('a', 4), ('i', 7)])

    def test_cross_map(self):
        """
        Tests the caae where only values are sent during crosses
        """
        item_counts = self.items.count()

        total = self.items \
                .a_group_by(lambda x: 1, lambda x: 1) \
                    .sum() \
                .map(lambda x: float(x[1]))

        results = item_counts \
                .cross_right(total, lambda ic, t: (ic[0], ic[1] / t)) \
                .read()

        results.sort()

        self.assertEquals(results, [(i , 1 / float(10)) for i in range(10, 20)])

    def test_len(self):
        """
        Tests the number of items in a collection.
        """

        self.assertEquals(self.items.len().read(), [10])
        self.assertEquals(Dampr.memory([]).len().read(), [0])

    def test_read_input(self):
        """
        Tests that custom taps work as expected.
        """
        class RangeDataset(Dataset):
            def __init__(self, n):
                self.n = n

            def read(self):
                for i in range(self.n):
                    yield i, i

        results = Dampr.read_input(RangeDataset(5), RangeDataset(10)) \
                .fold_by(lambda x: 1, lambda x, y: x + y) \
                .read()

        self.assertEqual(results[0][1], sum(range(5)) + sum(range(10)))

    def test_read_input(self):
        """
        Tests that custom taps work as expected.
        """
        class RangeDataset(Dataset):
            def __init__(self, n):
                self.n = n

            def read(self):
                for i in range(self.n):
                    yield i, i

        results = Dampr.read_input(RangeDataset(5), RangeDataset(10)) \
                .fold_by(lambda x: 1, lambda x, y: x + y) \
                .read()

        self.assertEqual(results[0][1], sum(range(5)) + sum(range(10)))

    def test_read_url(self):
        """
        Tests that we can read urls.
        """
        results = Dampr.read_input(UrlsInput(["http://www.example.com"])) \
                .filter(lambda line: 'h1' in line) \
                .map(lambda line: line.strip()) \
                .read()

        self.assertEqual(results, ['<h1>Example Domain</h1>'])

    def test_file_glob(self):
        """
        Tests that we can read file globs
        """
        import os
        files = []
        for i in range(10):
            path = os.path.join('/tmp', '_test_dampr_{}'.format(i))
            with open(path, 'w') as out:
                out.write(str(i))

            files.append(path)

        results = Dampr.text("/tmp/_test_dampr_[135]") \
                .map(int) \
                .fold_by(lambda x: 1, lambda x,y: x + y) \
                .read()

        self.assertEqual(results, [(1, 1 + 3 + 5)])

        for fname in files:
            os.unlink(fname)

    def test_top_k(self):
        """
        Tests getting the top k items
        """

        word = Dampr.memory(["supercalifragilisticexpialidociousa"])
        topk = word.flat_map(lambda w: list(w)).count() \
                .topk(5, lambda x: x[1])

        results = sorted(list(topk.run()))
        self.assertEquals(results, [('a', 4), ('c', 3), ('i', 7), ('l', 3), ('s', 3)])

if __name__ == '__main__':
    unittest.main()



import math
import os
import shutil
import logging
import multiprocessing
from multiprocessing import Queue
from multiprocessing.queues import Empty
import tempfile

from .base import *
from .dataset import *

CPUS = multiprocessing.cpu_count()

class Source(object):
    CNT = 0
    def __init__(self, name):
        self.name = name
        self.cnt = self.CNT
        type(self).CNT += 1

    def __hash__(self):
        return hash(self.cnt)

    def __eq__(self, other):
        return self.cnt == other.cnt

    def __str__(self):
        return "Source[`{}`]".format(self.name)

    __repr__ = __str__

class GMap(object):
    def __init__(self, output, inputs, mapper, combiner, shuffler, options=None):
        self.output = output
        self.inputs = inputs
        self.mapper = mapper
        self.combiner = combiner
        self.shuffler = shuffler
        self.options = options if options is not None else  {}

    def __unicode__(self):
        return u"Map".format()

    __repr__ = __unicode__

class GReduce(object):
    def __init__(self, output, inputs, reducer, options):
        self.output = output
        self.inputs = inputs
        self.reducer = reducer
        self.options = options if options is not None else  {}

    def __unicode__(self):
        return u"Reducer".format()

    __repr__ = __unicode__

class GSink(object):
    def __init__(self, output, inputs, mapper, path, options=None):
        self.output = output
        self.inputs = inputs
        self.mapper = mapper
        self.path = path
        self.options = options if options is not None else  {}

    def __unicode__(self):
        return u"Sink[path={}]".format(self.path)

    __repr__ = __unicode__

class Graph(object):
    def __init__(self):
        self.inputs = {}
        self.stages = []

    def _copy_graph(self):
        graph = Graph()
        graph.inputs.update(self.inputs)
        graph.stages.extend(self.stages)
        return graph

    def add_input(self, dataset):
        ng = self._copy_graph()
        inp = Source('Input:{}'. format(len(self.inputs)))
        ng.inputs[inp] = dataset
        return inp, ng

    def add_mapper(self, inputs, mapper, combiner=None, 
            shuffler=None, name=None, options=None):
        assert isinstance(mapper, Mapper)
        assert isinstance(combiner, (type(None), Combiner))
        assert isinstance(shuffler, (type(None), Shuffler))
        assert all(isinstance(inp, Source) for inp in inputs)
        if name is None:
            name = 'Map: {}'

        inp = Source(name.format(len(self.stages)))
        ng = self._copy_graph()
        ng.stages.append(GMap(inp, inputs, mapper, combiner, shuffler, options))
        return inp, ng

    def add_reducer(self, inputs, reducer, name=None, options=None):
        assert isinstance(reducer, Reducer)
        assert all(isinstance(inp, Source) for inp in inputs)
        if name is None:
            name = 'Reduce: {}'

        inp = Source(name.format(len(self.stages)))
        ng = self._copy_graph()
        ng.stages.append(GReduce(inp, inputs, reducer, options))
        return inp, ng

    def add_sink(self, inputs, mapper, path, name=None, options=None):
        assert isinstance(mapper, Mapper)
        assert all(isinstance(inp, Source) for inp in inputs)
        if name is None:
            name = 'Sink: {}'

        inp = Source(name.format(path))
        ng = self._copy_graph()
        ng.stages.append(GSink(inp, inputs, mapper, path, options))
        return inp, ng

    def union(self, other_graph):
        ng = self._copy_graph()
        ng.inputs.update(other_graph.inputs)
        seen_stages = set(ng.stages)
        for s in other_graph.stages:
            if s not in seen_stages:
                ng.stages.append(s)

        return ng
        
class RunnerBase(object):
    def __init__(self, name, graph, working_dir='/tmp'):
        self.file_system = FileSystem(os.path.join(working_dir, name))
        self.graph = graph

    def run_map(self, stage_id, data, mapper):
        raise NotImplementedError()

    def run_reduce(self, stage_id, data, reducer):
        raise NotImplementedError()

    def run_sink(self, stage_id, data, reducer):
        raise NotImplementedError()

    def format_outputs(self, outputs):
        new_ret = []
        for output in outputs:
            if len(output) == 1:
                output = output[0]
            else:
                output = MergeDataset(output)

            new_ret.append(output)
        
        return new_ret
        
    def collapse_datamappings(self, data_mappings):
        new_data = {}
        for dm in data_mappings:
            for partition, datasets in dm.items():
                if partition not in new_data:
                    new_data[partition] = []

                new_data[partition].extend(datasets)

        return new_data

    def run(self, outputs, cleanup=True):
        data = self.graph.inputs.copy()
        to_delete = set()
        splitter = Splitter()
        for stage_id, stage in enumerate(self.graph.stages):
            logging.info("Stage %s/%s", stage_id + 1, len(self.graph.stages))
            logging.info("Function - %s", stage)
            input_data = [data[i] for i in stage.inputs]
            for i, id in enumerate(input_data):
                logging.info("Input: %s", stage.inputs[i])

            logging.info("Output: %s", stage.output)

            cleanup_stage = True
            if isinstance(stage, GMap):
                data_mapping = self.run_map(stage_id, input_data, stage)

            elif isinstance(stage, GReduce):
                data_mapping = self.run_reducer(stage_id, input_data, stage)

            elif isinstance(stage, GSink):
                data_mapping = self.run_sink(stage_id, input_data, stage)
                # We're sinking to disk, don't delete it
                cleanup_stage = False

            else:
                raise TypeError("Unknown type")

            assert isinstance(data_mapping, dict)
            data[stage.output] = data_mapping
            if cleanup_stage:
                to_delete.add(stage.output)

        # Collect the outputs and determine what to delete
        ret = []
        for source in outputs:
            dataset = data[source]
            if isinstance(dataset, Dataset):
                cd = [dataset]
            elif isinstance(dataset, Chunker):
                cd = list(dataset.chunks())
            else:
                cd = [v for vs in dataset.values() for v in vs]

            ret.append(cd)
            if source in to_delete:
                to_delete.remove(source)

        ret = self.format_outputs(ret)
        # Cleanup
        if cleanup:
            for sd in to_delete:
                for ds in data[sd].values():
                    for d in ds:
                        d.delete()

        logging.info("Finished...")

        return ret


class StageRunner(object):
    def __init__(self, max_procs):
        self.max_procs = max_procs

    def execute_stage(self, t_id, payload):
        raise NotImplementedError()

    def run(self, job_queue):
        output_q = Queue()
        jobs = {}
        finished = []
        def get_output():
            child_pid, task_id, payload = output_q.get()
            finished.append(payload)
            
            # Cleanup pid
            jobs[child_pid].join()
            del jobs[child_pid]

        # Assign tasks via forking
        jobs_left = True
        while jobs_left:
            next_job = next(job_queue, None)
            jobs_left = next_job is not None
            if next_job is not None:
                # Are we full?  If so, wait for one to exit
                if len(jobs) == self.max_procs:
                    get_output()

                proc = self.execute_stage(next_job[0], next_job, output_q)
                proc.start()
                jobs[proc.pid] = proc

        while len(jobs) != 0:
            get_output()

        return finished

class MapStageRunner(StageRunner):
    def __init__(self, max_procs, fs, n_partitions, mapper, options):
        super(MapStageRunner, self).__init__(max_procs)
        self.fs = fs
        self.n_partitions = n_partitions
        self.mapper = mapper
        self.options = options

    def simple_map(self, job, out_q, fs):
        w_id = os.getpid()

        # Default job, nothing special
        if self.options.get('memory', False):
            writer_cls = BufferedSortedMemoryWriter
        else:
            writer_cls = BufferedSortedDiskWriter

        dw = CSDatasetWriter(fs, Splitter(), self.n_partitions, writer_cls=writer_cls)

        dw.start()
        m_id, main, supplemental = job
        logging.debug("Mapper %i: Computing map: %i", w_id, m_id)
        for k, v in self.mapper.mapper.map(main, *supplemental):
            dw.add_record(k, v)

        out_q.put((w_id, m_id, dw.finished()))
        logging.debug("Mapper: %i: Finished", w_id)

    def medium_map(self, job, out_q, combiner, shuffler, fs):
        """
        Runs a more fine grained map/combine/shuffler
        """
        w_id = os.getpid()
        if self.options.get('memory', False):
            dw = BufferedSortedMemoryWriter(fs)
        else:
            dw = BufferedSortedDiskWriter(fs)

        # Do we have a map side partial reducer?
        binop = self.options.get('binop')
        if callable(binop):
            reduce_buffer = self.options.get('reduce_buffer', 1000)
            if reduce_buffer > 0:
                # Zero buffer means all reductions will happen reduce side
                dw = ReducedWriter(dw, binop, max_values=reduce_buffer)

        dw.start()
        m_id, main, supplemental = job
        logging.debug("Mapper %i: Computing map: %i", w_id, m_id)
        for k, v in self.mapper.mapper.map(main, *supplemental):
            dw.add_record(k, v)

        sources = dw.finished()[0]

        if len(sources) > 1:
            logging.debug("Combining outputs: found %i files", len(sources))
            logging.debug("Combining: %s", sources)
            combined_stream = combiner.combine(sources)
        elif len(sources) == 1:
            combined_stream = sources[0]
        else:
            combined_stream = EmptyDataset()

        results = shuffler.shuffle(fs, [combined_stream])

        out_q.put((w_id, m_id, results))
        logging.debug("Mapper: %i: Finished", w_id)


    def execute_stage(self, t_id, payload, output_q):
        fs = self.fs.get_worker('map/{}'.format(t_id))

        if self.mapper.combiner is None and self.mapper.shuffler is None:

            p = multiprocessing.Process(target=self.simple_map,
                args=(payload, output_q, fs))
        else:
            c = NoopCombiner() if self.mapper.combiner is None else self.mapper.combiner
            if self.options.get('memory', False):
                writer_cls = UnorderedDiskWriter
            else:
                writer_cls = UnorderedMemoryWriter

            s = DefaultShuffler(self.n_partitions, Splitter(), writer_cls)
            o = self.mapper.options
            p = multiprocessing.Process(target=self.medium_map,
                args=(payload, output_q, c, s, fs))

        return p

class SinkStageRunner(StageRunner):
    def __init__(self, max_procs, mapper, path):
        super(SinkStageRunner, self).__init__(max_procs)
        if not os.path.exists(path):
            try:
                os.makedirs(path)
            except OSError as e:
                if e.errno != 17:
                    raise

        self.path = path
        self.mapper = mapper

    def sink(self, job, out_q):
        """
        Writes line delimited items as a sink
        """
        w_id = os.getpid()

        m_id, main, supplemental = job
        dw = SinkWriter(self.mapper.path, m_id)

        dw.start()
        logging.debug("Sink %i: Computing map: %i", w_id, m_id)
        for k, v in self.mapper.mapper.map(main, *supplemental):
            dw.add_record(k, v)

        out_q.put((w_id, m_id, dw.finished()))
        logging.debug("Sink: %i: Finished", w_id)

    def execute_stage(self, t_id, payload, output_q):
        p = multiprocessing.Process(target=self.sink,
            args=(payload, output_q))

        return p

class CombinerStageRunner(StageRunner):
    """
    Merges files together to reduce their on disk presence
    """
    def __init__(self, max_procs, fs, combiner, options):
        super(CombinerStageRunner, self).__init__(max_procs)
        self.fs = fs
        self.combiner = combiner
        self.options = options

    def combine(self, payload, output_q, fs):
        w_id = os.getpid()
        t_id, datasets = payload

        if self.options.get('memory', False):
            dw = ContiguousMemoryWriter(fs)
        else:
            dw = ContiguousDiskWriter(fs)

        dw.start()
        for k,v in self.combiner.combine(datasets):
            dw.add_record(k,v)

        for d in datasets:
            d.delete()

        output_q.put((w_id, t_id, dw.finished()[0]))

    def execute_stage(self, t_id, payload, output_q):
        fs = self.fs.get_worker('merge/{}'.format(t_id))

        p = multiprocessing.Process(target=self.combine,
            args=(payload, output_q, fs))

        return p

class ReduceStageRunner(StageRunner):
    def __init__(self, max_procs, fs, reducer, options):
        super(ReduceStageRunner, self).__init__(max_procs)
        self.fs = fs
        self.reducer = reducer
        self.options = options

    def reduce(self, job, out_q, dw):
        w_id = os.getpid()
        dw.start()
        r_id, datasets = job
        logging.debug("Reducer %i: Computing reduce: %i", w_id, r_id)
        for k, v in self.reducer.reducer.reduce(*datasets):
            dw.add_record(k, v)

        out_q.put((w_id, r_id, dw.finished()))
        logging.debug("Reducer: %i: Finished", w_id)

    def execute_stage(self, t_id, payload, output_q):
        fs = self.fs.get_worker('red/{}'.format(t_id))

        if self.options.get('memory', False):
            dw = ContiguousMemoryWriter(fs)
        else:
            dw = ContiguousDiskWriter(fs)

        m = multiprocessing.Process(target=self.reduce,
            args=(payload, output_q, dw))

        return m

class MTRunner(RunnerBase):
    def __init__(self, name, graph, 
            n_maps=CPUS, 
            n_reducers=CPUS,
            n_partitions=91,
            max_files_per_stage=50):
        super(MTRunner, self).__init__(name, graph)
        self.n_maps = n_maps
        self.n_reducers = n_reducers
        self.n_partitions = n_partitions
        self.max_files_per_stage = max_files_per_stage

    def chunk_list(self, v):
        """
        Given a list of files, creates subsets of those files.
        """
        chunks = min(self.max_files_per_stage, self.n_maps)
        num_files = min(int(math.ceil(len(v) / float(chunks))), self.max_files_per_stage)
        return ((i, v[s:s+num_files])
                for i, s in enumerate(range(0, len(v), num_files)))

    def run_map(self, stage_id, data_mappings, mapper):
        # if we get more than two input mappings, we only iterate over the first one
        iter_dm = data_mappings[0]
        if not isinstance(iter_dm, Chunker):
            iter_dm = DMChunker(iter_dm)

        supplementary = []
        for dm in data_mappings[1:]:
            if not isinstance(dm, Chunker):
                dm = DMChunker(dm)

            supplementary.append(list(dm.chunks()))

        jobs_queue = ((i, chunk, supplementary) for 
                i, chunk in enumerate(iter_dm.chunks()))

        stage_fs = self.file_system.get_stage(stage_id)
        n_maps = mapper.options.get('n_maps', self.n_maps)
        msr = MapStageRunner(n_maps, stage_fs, self.n_partitions, mapper, mapper.options)

        finished = msr.run(jobs_queue)

        collapsed = self.collapse_datamappings(finished)
        # Check for number of files
        for k, v in collapsed.items():
            while len(v) > self.max_files_per_stage:
                logging.debug("Partition %s needs to be merged: found %s files", k, len(v))
                chunks = self.chunk_list(v)
                c = NoopCombiner() if mapper.combiner is None else mapper.combiner
                csr = CombinerStageRunner(n_maps, stage_fs, c, mapper.options)
                v = [f for fs in csr.run(chunks) for f in fs]
                collapsed[k] = v

        return collapsed

    def run_reducer(self, stage_id, data_mappings, reducer):
        # Collect across inputs
        keys = sorted({k for dm in data_mappings for k in dm})
        transpose = {k: [] for k in keys}
        for dm in data_mappings:
            for k in keys:
                transpose[k].append(dm.get(k, []))
        
        stage_fs = self.file_system.get_stage(stage_id)
        n_reducers = reducer.options.get('n_reducers', self.n_reducers)
        rds = ReduceStageRunner(n_reducers, stage_fs, reducer, reducer.options)
        finished = rds.run(iter(transpose.items()))

        return self.collapse_datamappings(finished)

    def run_sink(self, stage_id, data_mappings, sink):
        # if we get more than two input mappings, we only iterate over the first one
        iter_dm = data_mappings[0]
        if not isinstance(iter_dm, Chunker):
            iter_dm = DMChunker(iter_dm)

        jobs_queue = ((i, chunk, data_mappings[1:]) 
                for i, chunk in enumerate(iter_dm.chunks()))
        n_maps = sink.options.get('n_maps', self.n_maps)
        ssr = SinkStageRunner(self.n_maps, sink, sink.path)

        finished = ssr.run(jobs_queue)

        return self.collapse_datamappings(finished)

    def format_outputs(self, outputs):
        
        # if mapped, we ordered the output so we use a merged dataset
        ret = []
        for output in outputs:
            while len(output) > self.max_files_per_stage:
                stage_fs = self.file_system.get_stage('final_combine')

                logging.debug("Combining final files: found %i", len(output))
                c = NoopCombiner() 
                csr = CombinerStageRunner(self.n_maps, stage_fs, c, {})
                jobs = self.chunk_list(output)
                output = [p for ps in csr.run(jobs) for p in ps]

            if len(output) == 1:
                output = output[0]

            else:
                output = MergeDataset(output)

            ret.append(output)

        return ret



"""
Dampr is a light-weight MapReduce library for single machine computation.  It supports a number of 
useful features such as map and reduce side joins, associative reduces, 
aggregations, multiprocessing, and more.  It operates by default out-of-core, allowing
processing of 100s of gbs to tbs of data on a single machine.

While the underlying engine uses the MapReduce paradigm for processing, Dampr is 
best utilized via it's DSL which provides higher level functionality for complex workflows.

It loosely attempts to replicate interfaces such as Spark or Scalding for ease of pickup.
"""
import logging
from .dampr import Dampr, PMap, PReduce, PJoin, ARReduce
from .base import BlockMapper, BlockReducer
from .dataset import Dataset

__all__ = ["Dampr", "PMap", "PReduce", "PJoin", "ARReduce", 
        "BlockMapper", "BlockReducer"]


def setup_logging(debug=False):
    """
    Convenience function for enabling logging
    """
    loglevel = logging.DEBUG if debug else logging.INFO
    logging.basicConfig(level=loglevel,
                        format='%(asctime)s %(levelname)s %(message)s')



from __future__ import print_function
import gzip
from math import ceil
import os
import sys
import itertools
import heapq
from operator import itemgetter

PY_MAJOR = sys.version_info.major

try:
    import cPickle as pickle
    from  cStringIO import StringIO

    def dump_pickle(obj, f):
        pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)

except ImportError:
    import pickle
    from io import BytesIO as StringIO

    def dump_pickle(o, f):
        pickle.dump(o, f,  pickle.HIGHEST_PROTOCOL)


class DatasetWriter(object):
    def __init__(self, worker_fs):
        self.worker_fs = worker_fs

    def start(self):
        raise NotImplementedError()

    def add_record(self, key, value):
        raise NotImplementedError()

    def finished(self):
        raise NotImplementedError()


class BufferedWriter(object):
    def __init__(self, f, max_size=1024**2):
        self.f = f
        self.max_size = max_size
        self.buffer = []
        self.size = 0
        self.empty = True

    def write(self, data):
        self.empty = False
        self.buffer.append(data)
        self.size += len(data)
        if self.size > self.max_size:
            self.flush()

    def flush(self):
        self.f.write(b''.join(self.buffer))
        self.f.flush()
        del self.buffer[:]
        self.size = 0

    def close(self):
        self.flush()
        self.f.close()
 
class ReducedWriter(DatasetWriter):
    """
    Used when a binary reducer function can partially join records on
    the Map side.  Writes out to an underlying DataWriter when the number
    of unique values in active reduction exceed a certain threshold
    """
    SENTINEL = object()
    def __init__(self, dw, binop, max_values):
        self.dw = dw
        self.binop = binop
        self.max_values = max_values
        self.key_set = {}

    def start(self):
        self.dw.start()
        self.key_set.clear()

    def add_record(self, key, value):
        cached_value = self.key_set.get(key, self.SENTINEL)
        if cached_value is not self.SENTINEL:
            self.key_set[key] = self.binop(value, cached_value)
        else:
            self.key_set[key] = value

            if len(self.key_set) > self.max_values:
                self.flush()

    def flush(self):
        for k, v in self.key_set.items():
            self.dw.add_record(k, v)

        self.key_set.clear()

    def finished(self):
        self.flush()
        return self.dw.finished()

class BufferedSortedWriter(DatasetWriter):
    """
    Used for most Map writers.  Takes key value pairs, serializes them
    with Pickle, and writes them to an internal buffer.  When the internal
    buffer exceeds the provided space, it flushes it to a new file on disk.
    """
    def __init__(self, fs, buffer_size=1*1024**2, always_to_disk=True):
        self.fs = fs
        self.buffer_size = buffer_size
        self.always_to_disk = always_to_disk

    def start(self):
        self.files = []
        self.buffer = StringIO()
        self.keyoffs = []

    def _write_to_gzip(self, f):
        with gzip.GzipFile(fileobj=f, mode='wb', compresslevel=1) as f:
            bw = BufferedWriter(f)
            for kv in self._sort_kvs():
                bw.write(kv)

            bw.flush()

    def _sort_kvs(self):
        self.keyoffs.sort(key=itemgetter(0))
        v = self.buffer.getvalue()
        for _, start, stop in self.keyoffs:
            yield v[start:stop]

    def flush(self):
        raise NotImplementedError()

    def _write_to_buffer(self, key, value):
        key_start = self.buffer.tell()
        dump_pickle((key, value), self.buffer)
        return key_start, self.buffer.tell()

    def add_record(self, key, value):
        kvs, kvst = self._write_to_buffer(key, value)
        self.keyoffs.append((key, kvs, kvst))
        if self.buffer.tell() > self.buffer_size:
            self._flush_buffer()

    def reset(self):
        self.buffer.truncate(0)
        self.buffer.seek(0)
        self.keyoffs = []

    def _flush_buffer(self):
        dataset = self.flush()
        self.files.append(dataset)
        self.reset()

    def finished(self):
        if self.buffer.tell() > 0:
            self._flush_buffer()

        return {0: self.files}

class BufferedSortedMemoryWriter(BufferedSortedWriter):
    """
    Writes out the Buffer to Memory
    """
    def __init__(self, fs, buffer_size=1*1024**2):
        self.fs = fs
        self.buffer_size = buffer_size

    def flush(self):
        sio = StringIO()
        self._write_to_gzip(sio)
        return MemGZipDataset(sio.getvalue())

class BufferedSortedDiskWriter(BufferedSortedWriter):
    """
    Writes out Buffer to Disk
    """
    def write_batch_to_disk(self):
        path = self.fs.get_file(str(len(self.files)))
        with open(path, 'wb') as out:
            self._write_to_gzip(out)

        return path

    def flush(self):
        file_name = self.write_batch_to_disk()
        return PickledDataset(file_name)

class CSDatasetWriter(DatasetWriter):
    """
    Writes out a stream of key values, splitting into separate partitions based on
    the provided Splitter.
    """
    def __init__(self, worker_fs, splitter, n_partitions, 
            buffer_size=10*1024*1024, writer_cls=BufferedSortedDiskWriter):
        super(CSDatasetWriter, self).__init__(worker_fs)
        self.splitter     = splitter
        self.n_partitions = n_partitions
        self.buffer_size  = buffer_size

        assert issubclass(writer_cls, BufferedSortedWriter)
        self.writer_cls = writer_cls

    def start(self):
        self.partitions = []
        for i in range(self.n_partitions):
            sub_fs = self.worker_fs.get_substage('partition_{}'.format(i))
            self.partitions.append(self.writer_cls(sub_fs, self.buffer_size))
            self.partitions[-1].start()

    def add_record(self, key, value):
        nidx = self.splitter.partition(key, self.n_partitions)
        self.partitions[nidx].add_record(key, value)

    def finished(self):
        return {i: p.finished()[0] for i, p in enumerate(self.partitions)}

class SinkWriter(DatasetWriter):
    """
    Writes out the value portion of the key/value pair in text format.
    """
    def __init__(self, path, idx):
        super(SinkWriter, self).__init__(None)
        self.path = path
        self.idx = idx
        self.fname = os.path.join(self.path, str(self.idx))

    def start(self):
        self.f = open(self.fname, 'w')

    def add_record(self, key, value):
        print(value, file=self.f)

    def finished(self):
        self.f.close()
        return {0: [TextLineDataset(self.fname)]}

class ContiguousWriter(DatasetWriter):
    """
    Writes out data unordered into a Gzipped file
    """
    def __init__(self, worker_fs, batch_size=10):
        super(ContiguousWriter, self).__init__(worker_fs)
        self.worker_fs = worker_fs
        self.batch_size = batch_size
    
    def get_fileobj(self):
        raise NotImplementedError()

    def get_dataset(self):
        raise NotImplementedError()
    
    def start(self):
        self.buffer = []
        self.f = BufferedWriter(self.get_fileobj())
    
    def add_record(self, key, value):
        self.buffer.append((key, value))
        if len(self.buffer) == self.batch_size:
            self.flush()

    def flush(self):
        dump_pickle(self.buffer, self.f)
        self.buffer = []

    def finished(self):
        self.flush()
        if self.f.empty:
            self.f.close()
            return {0: []}

        self.f.close()
        return {0: [self.get_dataset()]}

class ContiguousDiskWriter(ContiguousWriter):
    """
    Writes out data unordered into a Gzipped file
    """
    def get_fileobj(self):
        self.fname = self.worker_fs.get_file()
        return gzip.GzipFile(self.fname, 'wb', 1)

    def get_dataset(self):
        return PickledDataset(self.fname, True)
    
class ContiguousMemoryWriter(ContiguousWriter):
    """
    Writes out data unordered into a Gzipped file
    """
    def get_fileobj(self):
        self.buffer = StringIO()
        return gzip.GzipFile(fileobj=self.buffer, mode='wb', compresslevel=1)

    def get_dataset(self):
        return MemGZipDataset(self.buffer.getvalue(), True)

class UnorderedWriter(DatasetWriter):
    """
    Writes unordered chunks to disk
    """
    def __init__(self, worker_fs, batch_size=10, chunk_size=64*1024*1024, always_to_disk=False):
        super(UnorderedWriter, self).__init__(worker_fs)
        self.chunk_size = chunk_size
        self.chunk_id = 0
        self.batch_size = batch_size
        self.files = []
        self.always_to_disk = always_to_disk
    
    def _write_to_gzip(self, fobj):
        buf = self.buffer
        with gzip.GzipFile(fileobj=fobj, mode='wb', compresslevel=1) as f:
            buf.seek(0)
            data = buf.read(4096 * 4)
            while data:
                f.write(data)
                data = buf.read(4096 * 4)

    def flush(self):
        raise NotImplementedError()

    def flush_batch(self):
        if len(self.batch) > 0:
            dump_pickle(self.batch, self.buffer)
            self.batch = []

    def reset(self):
        self.buffer = StringIO()

    def start(self):
        self.batch = []
        self.reset()

    def add_record(self, key, value):
        self.batch.append((key, value))
        if len(self.batch) == self.batch_size:
            self.flush_batch()
            if self.buffer.tell() > self.chunk_size:
                self._flush_buffer()

    def _flush_buffer(self):
        dataset = self.flush()
        self.files.append(dataset)
        self.reset()

    def finished(self):
        self.flush_batch()
        if self.buffer.tell() > 0:
            self._flush_buffer()

        return {0: self.files}

class UnorderedDiskWriter(UnorderedWriter):
    """
    Writes unordered chunks to disk
    """
    def flush(self):
        chunk_name = self.worker_fs.get_file(str(self.chunk_id))
        self.chunk_id += 1
        with open(chunk_name, 'wb') as out:
            self._write_to_gzip(out)
        
        return PickledDataset(chunk_name, True)

class UnorderedMemoryWriter(UnorderedWriter):
    """
    Writes unordered chunks to memory
    """
    def flush(self):
        tmp = StringIO()
        self._write_to_gzip(tmp)
        # Compress the data 
        return MemGZipDataset(tmp.getvalue(), True)

class Chunker(object):
    def chunks(self):
        raise NotImplementedError()

class Dataset(Chunker):

    def read(self):
        raise NotImplementedError()

    def grouped_read(self):
        for key, group in itertools.groupby(self.read(), key=lambda x: x[0]):
            group = list(group)
            it = (kv[1] for kv in group)
            yield key, it

    def delete(self):
        raise NotImplementedError()

    def __iter__(self):
        return self.read()

    def chunks(self):
        yield self

class EmptyDataset(Dataset):

    def read(self):
        return iter([])

    def delete(self):
        pass

class TextLineDataset(Dataset):
    def __init__(self, path, start=0, end=None):
        self.path = path
        self.start = start
        self.end = end

    def read(self):
        with open(self.path) as f:
            f.seek(self.start)
            cur_pos = self.start
            if self.start > 0:
                cur_pos += len(f.readline())

            for i, line in enumerate(f):
                yield self.start + i, line.rstrip(os.linesep)
                cur_pos += len(line)
                if self.end is not None and cur_pos > self.end:
                    break

    def __str__(self):
        return "Text[path={},start={},end={}]".format(self.path,self.start, self.end)

    def delete(self):
        pass

class PickledDataset(Dataset):
    def __init__(self, path, batched=False):
        self.path = path
        self.batched = batched

    def read(self):
        with gzip.GzipFile(self.path, 'rb', 1) as f:
            try:
                if self.batched:
                    while True:
                        for d in pickle.load(f):
                            yield d
                else:
                    while True:
                        yield pickle.load(f)

            except EOFError:
                pass

    def delete(self):
        os.unlink(self.path)

    def __str__(self):
        return 'PickledDataset[path={}]'.format(self.path)
    __repr__ = __str__


class MemGZipDataset(Dataset):
    def __init__(self, sio, batched=False):
        self.sio = sio
        self.batched = batched

    def read(self):
        with gzip.GzipFile(fileobj=StringIO(self.sio)) as sio:
            try:
                if not self.batched:
                    while True:
                        yield pickle.load(sio)
                else:
                    while True:
                        for d in pickle.load(sio):
                            yield d
            except EOFError:
                pass

    def delete(self):
        pass

    
class CatDataset(Dataset, Chunker):
    def __init__(self, datasets):
        self.datasets = datasets

    def read(self):
        for ds in self.datasets:
            for x in ds.read():
                yield x

    def delete(self):
        for d in self.datasets:
            d.delete()

    def chunks(self):
        for d in self.datasets:
            yield d

class MergeDataset(Dataset, Chunker):
    def __init__(self, datasets):
        self.datasets = datasets

    def read(self):
        if len(self.datasets) == 1:
            return self.datasets[0].read()

        its = [ds.read() for ds in self.datasets]
        if PY_MAJOR == 3:
            return heapq.merge(*its, key=lambda x: x[0])

        return heapq.merge(*its)

    def chunks(self):
        for d in self.datasets:
            yield d


    def delete(self):
        for d in self.datasets:
            d.delete()

class MemoryDataset(Dataset):
    def __init__(self, kvs, partitions=13):
        self.kvs = kvs
        self.partitions = partitions

    def read(self):
        for k, v in self.kvs:
            yield k, v

    def delete(self):
        pass

    def chunks(self):
        if self.partitions == 1:
            yield self
        else:
            chunk_size = int(ceil(len(self.kvs) / float(self.partitions)))
            start = 0
            while start < len(self.kvs):
                yield MemoryDataset(self.kvs[start:start+chunk_size], 1)
                start += chunk_size

class StreamDataset(Dataset):
    def __init__(self, it):
        self.it = it
    
    def read(self):
        return self.it

    def delete(self):
        pass

class DMChunker(Chunker):
    def __init__(self, data_mapping):
        self.dm = data_mapping

    def chunks(self):
        for vs in self.dm.values():
            for v in vs:
                yield v




try:
    from urllib.request import urlopen
    from urllib.error import HTTPError
except ImportError:
    from urllib2 import urlopen, HTTPError

import glob
from contextlib import closing
import os

from .dataset import Chunker, TextLineDataset, MemoryDataset, Dataset

class PathInput(Chunker):
    def __init__(self, path, chunk_size=64*1024**2):
        self.path = path
        self.chunk_size = chunk_size

    def chunks(self):
        for path in glob.glob(self.path):
            if os.path.isfile(path):
                for c in TextInput(path, self.chunk_size).chunks():
                    yield c
            else:
                for root, dirs, files in os.walk(self.directory):
                    for fname in files:
                        path = os.path.join(root, fname)
                        for chunk in TextInput(path, self.chunk_size).chunks():
                            yield chunk

class TextInput(Chunker):
    def __init__(self, path, chunk_size=64*1024**2):
        self.path = path
        self.chunk_size = chunk_size

    def chunks(self):
        file_size = os.stat(self.path).st_size
        offset = 0
        while offset < file_size:
            yield TextLineDataset(self.path, offset, offset + self.chunk_size)
            offset += self.chunk_size

class MemoryInput(Chunker):
    def __init__(self, items, partitions=50):
        self.items = items
        self.partitions = min(len(items), partitions)

    def chunks(self):
        # Memory Dataset could be zero
        if self.partitions == 0:
            yield MemoryDataset(self.items)
        else:
            chunk_size = int(len(self.items) // float(self.partitions))
            for start in range(0, len(self.items), chunk_size):
                yield MemoryDataset(self.items[start:start+chunk_size])

class UrlsInput(Chunker):
    def __init__(self, urls, skip_on_error=True):
        self.urls = urls
        self.soe = skip_on_error

    def chunks(self):
        for url in self.urls:
            yield UrlDataset(url, self.soe)

class UrlDataset(Dataset):
    def __init__(self, path, skip_on_error=True):
        self.path = path
        self.soe = skip_on_error

    def read(self):
        try:
            with closing(urlopen(self.path)) as h:
                for i, line in enumerate(h):
                    yield i, line.decode('utf-8')

        except HTTPError:
            if not self.soe:
                raise




import itertools
import sys
import operator
import time
import logging
import json
import random

from .base import *
from .runner import MTRunner, Graph, Source
from .dataset import Chunker, CatDataset
from .inputs import MemoryInput, PathInput

class ValueEmitter(object):
    """
    Reads values from a processed dataset.  Can be used on the shell
    to read the results of a computation.
    """
    def __init__(self, datasets):
        self.datasets = datasets

    def stream(self):
        """
        Streams the values in the dataset.
        """
        for _, v in self.datasets.read():
            yield v

    def read(self, k=None):
        """
        Reads the first k items from the dataset.  If k is None, reads the
        entire dataset.
        """
        if k is None:
            return list(self.stream())

        return list(itertools.islice(self.stream(), k))

    def __iter__(self):
        return self.stream()

    def delete(self):
        """
        Deletes the cached on-disk dataset.
        """
        self.datasets.delete()

class PBase(object):
    """
    Base Dampr class
    """
    def __init__(self, source, pmer):
        assert isinstance(source, Source)
        self.source = source
        self.pmer = pmer

    def run(self, name=None, **kwargs):
        """
        Evaluates the composed Dampr graph with the provided name and subsequent options.
        By default, uses /tmp as temporary storage.

        Returns a ValueEmitter useful for shell access.
        """
        if name is None:
            name = 'dampr/{}'.format(random.random())

        logging.debug("run source: %s", self.source)
        ds = self.pmer.runner(name, self.pmer.graph, **kwargs).run([self.source])
        return ValueEmitter(ds[0])

    def read(self, k=None, **kwargs):
        """
        Shorthand for run() followed by a read()
        """
        return self.run(**kwargs).read(k)

def _identity(k, v):
    yield k, v

class PMap(PBase):
    """
    Represents most mapping processes.  Internally, it collects consecutive mapping operations
    and fuses them together into a single operation, speeding up many types of operations.

    This class shouldn't be intialized manually.
    """

    def __init__(self, source, pmer, agg=None):
        super(PMap, self).__init__(source, pmer)
        self.agg = [] if agg is None else agg

    def run(self, name=None, **kwargs):
        """
        Run the defined graph.
        """
        if len(self.agg) > 0:
            return self.checkpoint().run(name, **kwargs)
        else:
            return super(PMap, self).run(name, **kwargs)

    def _add_mapper(self, mapper):
        assert isinstance(mapper, Streamable)
        # If we have a fusable mapper, just add it to the set.
        return PMap(self.source, self.pmer, self.agg + [mapper])

    def _add_map(self, f):
        return self._add_mapper(Map(f))

    def sample(self, prob):
        """
        Samples data with a given probability.  For example:
        
        `graph.sample(0.1)` will uniformly sample 10% of the data in the collection.
        """
        assert 0 <= prob <= 1.0

        def _sample(k, v):
            if get_rand().random() < prob:
                yield k, v

        return self._add_map(_sample)
        
    def checkpoint(self, force=False, combiner=None, options=None):
        """
        Checkpoint forces Dampr to fuse all cached maps and add it as a MR stage.

        This is useful when sharing the results of a computation with multiple other graphs.
        
        Without checkpoint(), Dampr would execute the shared graph multiple times rather than reuse
        the results of the computation:

            >>> evens = Dampr.memory([1,2,3,4,5]).filter(lambda x: x % 2 == 0).checkpoint()
            >>> summed = evens.group_by(lambda x: 1).sum()
            >>> multiplied = evens.group_by(lambda x: 1).reduce(lambda x, y: x * y)

        """
        if len(self.agg) > 0 or force:
            aggs = [Map(_identity)] if len(self.agg) == 0 else self.agg[:]
            name = ' -> ' .join('{}'.format(str(a)) for a in aggs)
            name = 'Stage {}: %s' % (name)
            source, pmer = self.pmer._add_mapper([self.source], 
                    fuse(aggs), 
                    combiner=combiner,
                    name=name,
                    options=options)
            return PMap(source, pmer) 

        return self

    def custom_mapper(self, mapper, name=None, **options):
        """
        Custom Mapper provides a low-level interface to the underlying map function.
        Users can provide any instance which adhere's to the Mapper interface, allowing
        for powerful or specific implementations.

        This is typically used very rarely and has some caveats such as not fusing
        with other Mappers.  Similarly, the implementer needs to understand more of the
        nuances associated with keys.

            >>> from dampr.base import Map
            >>> Dampr.memory([1,2,3,4,5]).custom_mapper(Map(lambda k, x: [(k, x+1)])).read()
            [2, 3, 4, 5, 6]
        """
        if isinstance(mapper, Streamable):
            return self._add_mapper(mapper)

        assert isinstance(mapper, Mapper)
        name = name if name is not None else str(mapper)
        me = self.checkpoint()
        source, pmer = me.pmer._add_mapper([me.source], 
                mapper,
                name=name,
                options=options)
        
        return PMap(source, pmer)

    def custom_reducer(self, reducer, name=None, **options):
        """
        Allows the user to provide any Reducer which adheres to the Reducer interface.
        This is a very powerful, low-level interface which should be avoided when possible
        as it's easy to write bugs.

            >>> Dampr.memory([1,2,3,4,5]).custom_reducer(Reduce(lambda k, x: [(k, sum(x))])).read()
            [[(0, 1)], [(1, 2)], [(2, 3)], [(3, 4)], [(4, 5)]]
        """
        assert isinstance(reducer, Reducer)
        me = self.checkpoint(force=True)
        name = name if name is not None else str(reducer)
        new_source, pmer = me.pmer._add_reducer([me.source], 
                reducer, 
                name=name,
                options=options)
        
        return PMap(new_source, pmer)

    def partition_map(self, f, **options):
        """
        Provides a medium-level interface for writing custom functionality for mapping 
        a partition.  This can be used for creating custom logic in mappers for the sake
        of performance or additional functionality.

        Note: partition_map will still execute even if the partition is empty!  Make sure
        the logic handles empty sets!

        `f` is a function that takes in an iterator of items in the partition and yields 
        out group keys and new values.


            >>> def plus_one(items):
            ...   for num in items:
            ...     yield num, num + 1
            ...
            >>> Dampr.memory([1,2,3,4,5]).partition_map(plus_one).read()
            [2, 3, 4, 5, 6]

        """
        return self.custom_mapper(StreamMapper(f), **options)

    def partition_reduce(self, f):
        """
        `partition_reduce` is a medium-level function that allows for more complex logic
        during reductions.  It can be useful in certain cases where reductions over sets
        of reduced values is convenient, such as returning only the top K items in a dataset

        Note: `partition_reduce` will still execute on empty partitions!  Make sure the
        logic handles cases where a partition is empty!

            >>> def largest_number(it):
            ...   largest = float('-inf')
            ...   for group_key, its in it:
            ...     for value in its:
            ...       largest = max(largest, value)
            ...   yield "Largest", largest
            ...
            >>> Dampr.memory([1,2,3,4,5]).partition_reduce(largest_number).read(n_partitions=1)
            [('Largest', 5)]
        """
        return self.custom_reducer(StreamReducer(f))

    def len(self):
        """
        Counts the number of items in the 
	collection.

	    >>> Dampr.memory([1,2,3,4,5]).len().read()
	    [5]
        """

        def _map_count(items):
            count = 0
            for _ in items:
                count += 1

            yield 1, count

        def _reduce_count(groups):
            count = 0
            not_empty = False
            for _, counts in groups:
                not_empty = True
                for c in counts:
                    count += c

            if not_empty:
                yield 1, count

        return self \
                .partition_map(_map_count) \
                .partition_reduce(_reduce_count) \
                .map(lambda x: x[1])

    def map(self, f):
        """
        Maps elements in the underlying collection using function 
        `f`:

            >>> Dampr.memory([1,2,3,4,5]).map(lambda x: x + 1).read()
            [2, 3, 4, 5, 6]
        """
        def _map(k, v):
            yield k, f(v)

        return self._add_map(_map)

    def filter(self, f):
        """
        Filters items from a collection based on a predicate f.
        A predicate return True keeps the item.

            >>> Dampr.memory([1,2,3,4,5]).filter(lambda x: x % 2 == 1).read()
            [1, 3, 5]

        """
        def _filter(k, v):
            if f(v):
                yield k, v

        return self._add_map(_filter)

    def flat_map(self, f):
        """
        Maps elements in the underlying collection using function f, 
        flattening the results

            >>> Dampr.memory([1,2,3,4,5]).flat_map(range).read()
            [0, 0, 1, 0, 1, 2, 0, 1, 2, 3, 0, 1, 2, 3, 4]
        """
        def _flat_map(k, v):
            for vi in f(v):
                yield k, vi

        return self._add_map(_flat_map)

    def group_by(self, key, vf=lambda x: x):
        """
        Groups a collections of X by a key function, optionally mapping X to Y 
        using `vf`.  Returns a Reducer object for different types of aggregations

            >>> Dampr.memory([1,2,3,4,5]).group_by(lambda x: x % 2).reduce(lambda k, it: sum(it)).read()
            [(0, 6), (1, 9)]
        """
        def _group_by(_key, value):
            yield key(value), vf(value)

        pm = self._add_map(_group_by).checkpoint()
        return PReduce(pm.source, pm.pmer)

    def a_group_by(self, key, vf=lambda x: x):
        """
        Groups a collection of X by a key function and optionally mapping X to Y
        using `vf`.  It differs from `group_by` by requiring an associative reduction
        operator: by forcing this restriction, Dampr is able to perform a partial 
        reduction of the collection during the mapping sequence which can dramatically
        speed up the performace of the reduce stage.

        When possible, use a_group_by over the more general group_by.

            >>> Dampr.memory([1,2,3,4,5]).a_group_by(lambda x: x % 2).reduce(lambda x, y: x+y).read()
            [(0, 6), (1, 9)]
        """
        def _a_group_by(_key, value):
            yield key(value), vf(value)

        # We don't checkpoint here!
        pm = self._add_map(_a_group_by)
        return ARReduce(pm)

    def fold_by(self, key, binop, value=lambda x: x, **options):
        """
        Shortcut for a_group_by(key, value).reduce(binop)
        """
        return self.a_group_by(key, value).reduce(binop, **options)

    def sort_by(self, key, **options):
        """
        Sorts the results by a given key function.
        
            >>> Dampr.memory([1,2,3,4,5]).filter(lambda x: x % 2 == 1).sort_by(lambda x: -x).read()
            [5, 3, 1]
        """
        def _sort_by(_key, value):
            yield key(value), value

        return self._add_map(_sort_by).checkpoint(options=options)

    def join(self, other):
        """
        Joins two independent computations, returning a Joining class.

        This is a powerful and expensive operation which can merge two Dampr 
        collections together.
        """
        assert isinstance(other, PBase)
        me = self.checkpoint(True)
        if isinstance(other, PMap):
            other = other.checkpoint(True)

        pmer = Dampr(me.pmer.graph.union(other.pmer.graph))
        return PJoin(me.source, pmer, other.source)

    def count(self, key=lambda x: x, **options):
        """
        Counts each item X in the collection by its key 
        function.

            >>> Dampr.memory([1,2,3,4,5]).count(lambda x: x % 2).read()
            [(0, 2), (1, 3)]
        """
        return self.a_group_by(key, lambda v: 1) \
                .reduce(operator.add, **options)

    def mean(self, key=lambda x: 1, value=lambda x: x, **options):
        """
        Finds the mean of X, grouped by its key function and optionally
        mapped by a value function:

            >>> ages = [("Andrew", 33), ("Alice", 42), ("Andrew", 12), ("Bob", 51)]
            >>> Dampr.memory(ages).mean(lambda x: x[0], lambda v: v[1]).read()
            [('Alice', 42.0), ('Andrew', 22.5), ('Bob', 51.0)]
        """
        def _mean_binop(x, y):
            return x[0] + y[0], x[1] + y[1]

        def _average(x):
            return (x[0], x[1][0] / float(x[1][1]))

        return self.a_group_by(key, lambda v: (value(v), 1)) \
                .reduce(_mean_binop, **options) \
                .map(_average)

    def inspect(self, prefix="", exit=False):
        """
        Inspect is a debug function which prints each item X that flows through
        it.  It's valueable in inspecting intermediate results of a pipeline 
        without changing the values internally.
        """
        def _inspect(k, v):
            print("{}: {}".format(prefix, v))
            yield k, v

        ins = self._add_map(_inspect)
        if exit:
            ins.run()
            sys.exit(0)

        return ins

    def cached(self, **options):
        """
        The cached function runs a graph and stores it in memory.  This is 
        useful for small datasets to gain extra performance in subsequent 
        computations.

            >>> Dampr.memory([1,2,3,4,5,6]).mean(lambda x: x % 2).cached().read()
            [(0, 4.0), (1, 3.0)]
        """
        # Run the pipeline, load it into memory, and create a new graph
        options['memory'] = True
        return self.checkpoint(options=options)

    def sink(self, path):
        """
        Since writes each X in a collection to a given path.  The path will
        create a directory and write each map or reduce partition into the
        directory as partitions.  Sink assumes each X in the collection is 
        already a unicode string.

            >>> Dampr.memory(["foo", "bar", "baz"]).sink("/tmp/foo").run()
            >>> open("/tmp/foo/0").read()
            >>> open("/tmp/foo/1").read()
            >>> open("/tmp/foo/2").read()
    """
        aggs = [Map(_identity)] if len(self.agg) == 0 else self.agg[:]
        name = ' -> ' .join('{}'.format(a.mapper.__name__) for a in aggs)
        name = 'Stage {}: %s' % (name)
        source, pmer = self.pmer._add_sink([self.source], 
                fuse(aggs), 
                path=path,
                name=name,
                options=None)
        return PMap(source, pmer) 

    def sink_tsv(self, path):
        """
        A convenience function which takes a tuple or list, creates a simple
        tab-delimited output, and sinks it to the provided path:

            >>> Dampr.memory([("Hank Aaron", 755)]).sink_tsv("/tmp/foo").run()
            >>> open("/tmp/foo/0").read()
        """
        return self.map(lambda x: u'\t'.join(unicode(p) for p in x)).sink(path)

    def sink_json(self, path):
        """
        A convenience function which takes a simple python object and serializes
        it to a line-delimited json to the given path:

            >>> Dampr.memory([{"name": "Hank Aaron", "home runs": 755}]).sink_json("/tmp/foo").run()
            >>> open("/tmp/foo/0").read()
        """
        return self.map(json.dumps).sink(path)

    def cross_right(self, other, cross, memory=False):
        """
        Produces the cross product between two datasets during the map stage.
        This is an incredibly expensive operation when done against two large
        datasets.  

        When it is known that the right dataset is small, setting `memory` to `True`
        will cache the right dataset in memory, speeding up the computation significantly.

        Two items, Xi and Yi, are joined with the cross function.

            >>> left = Dampr.memory([1,2,3,4,5])
            >>> right = Dampr.memory(['foo', 'bar'])
            >>> left.cross_right(right, lambda x, y: (x, y)).read()
            [(1, 'foo'), (1, 'bar'), (2, 'foo'), (2, 'bar'), (3, 'foo'), (3, 'bar'), (4, 'foo'), (4, 'bar'), (5, 'foo'), (5, 'bar')]
        """
        assert isinstance(other, PMap)
        return other.cross_left(self, lambda xi, yi: cross(yi, xi), memory)

    def cross_left(self, other, cross, memory=False, **options):
        """
        Produces the cross product between two datasets during the map stage.
        This is an incredibly expensive operation when done against two large
        datasets.  

        When it is known that the left dataset is small, setting `memory` to `True`
        will cache the right dataset in memory, speeding up the computation significantly.

        Two items, Xi and Yi, are joined with the cross function.

            >>> left = Dampr.memory([1,2,3,4,5])
            >>> right = Dampr.memory(['foo', 'bar'])
            >>> left.cross_left(right, lambda x, y: (x, y)).read()
            [(1, 'foo'), (2, 'foo'), (3, 'foo'), (4, 'foo'), (5, 'foo'), (1, 'bar'), (2, 'bar'), (3, 'bar'), (4, 'bar'), (5, 'bar')]
        """
        def _cross(k1, v1, k2, v2):
            yield k1, cross(v2, v1)

        me = self.checkpoint()
        other = other.checkpoint()
        pmer = Dampr(me.pmer.graph.union(other.pmer.graph))
        name = 'Stage {}: Cross'
        source, pmer = pmer._add_mapper([other.source, me.source], 
                MapCrossJoin(_cross, cache=memory), 
                combiner=None,
                name=name,
                options=options)
        return PMap(source, pmer) 

    def topk(self, k, value=None):
        """
        Computes the top k elements in the set.  The 'value' function is responsible for
        returning a comparable object.

            >>> Dampr.memory([1,3,2,4,2.2]).topk(2).read()
            [3, 4]
            >>> Dampr.memory([1,3,2,4,2.2]).topk(2, lambda x: -x).read()
            [1, 2]
        """
        if value is None:
            value = lambda x: x

        import heapq
        def map_topk(it):
            heap = []
            for x in it: 
                heapq.heappush(heap, (value(x), x))
                if len(heap) > k:
                    heapq.heappop(heap)

            return ((1, x) for x in heap)

        def reduce_topk(it):
            counts = (v for k, vit in it for v in vit)
            for count, x in heapq.nlargest(k, counts):
                yield x, 1

        return self \
                .partition_map(map_topk) \
                .partition_reduce(reduce_topk) \
                .map(lambda x: x[0])

class ARReduce(object):
    """
    Associative Reducer operators.
    """
    def __init__(self, pmap):
        self.pmap = pmap

    def reduce(self, binop, reduce_buffer=1000, **options):
        """
        Reduces a grouped dataset by an associative binary operator.
        This will do a partial reduce in the map stage, completing the reduction
        in the reduce stage.  It is often substantially faster than the more
        general group_by.

        `reduce_buffer` is a constant which tells Dampr how much temporary storage
        to keep in memory on the map side reductions.  For example with `1000`, Dampr
        will keep 1000 unique keys in memory.  In the case where a new key would spill
        over the buffer size, Dampr will flush the buffer to disk and create a new
        buffer.  By increasing the `reduce_buffer`, you can increase efficiency while
        sacrificing memory.

            >>> Dampr.memory([1,2,3,4,5]).a_group_by(lambda x: 1).reduce(lambda x, y: x + y).read()
            [(1, 15)]
        """
        def _reduce(key, vs):
            acc = next(vs)
            for v in vs:
                acc = binop(acc, v)

            return acc

        red = Reduce(_reduce)
        options.update({"binop": binop, "reduce_buffer": reduce_buffer})
        # We add the associative aggregator to the combiner during map
        pm = self.pmap.checkpoint(True, 
                combiner=PartialReduceCombiner(red), 
                options=options)
        return PReduce(pm.source, pm.pmer).reduce(_reduce)
    
    def first(self, **options):
        """
        Returns the first item found for a given key. 
            >>> Dampr.memory([1,2,3,4,5]).a_group_by(lambda x: x % 2).first().read()
            [(0, 2), (1, 1)]
        """
        return self.reduce(lambda x, _y: x, **options)

    def sum(self, **options):
        """
        Simple sum of values by key.

            >>> Dampr.memory([1,2,3,4,5]).a_group_by(lambda x: x % 2).sum().read()
            [(0, 6), (1, 9)]
        """
        return self.reduce(lambda x, y: x + y, **options)

    
class PReduce(PBase):
    """
    A more general reduce class.
    """

    def reduce(self, f):
        """
        Reduces a grouped set of items by f, which takes two arguments: the group key
        and an iterator lazily yield all items in the group.

            >>> Dampr.memory([1,2,3,4,5]).group_by(lambda x: x % 2).reduce(lambda k, it: sum(it)).read()
            [(0, 6), (1, 9)]
        """
        new_source, pmer = self.pmer._add_reducer([self.source], KeyedReduce(f))
        return PMap(new_source, pmer)

    def unique(self, key=lambda x: x):
        """
        Returns the unique set of items in the grouping.

            >>> names = [("Andrew", 1), ("Andrew", 1), ("Andrew", 2), ("Becky", 13)]
            >>> Dampr.memory(names).group_by(lambda x: x[0], lambda x: x[1]).unique().read()
            [('Andrew', [1, 2]), ('Becky', [13])]
        """
        def _uniq(k, it):
            seen = set()
            agg = []
            for v in it:
                fv = key(v)
                if fv not in seen:
                    seen.add(fv)
                    agg.append(v)

            return agg

        return self.reduce(_uniq)

    def join(self, other):
        """
        Performs a join between two datasets on a given key, returning a PJoin object.
        """
        assert isinstance(other, PBase)
        if isinstance(other, PMap):
            other = other.checkpoint(True)

        pmer = Dampr(self.pmer.graph.union(other.pmer.graph))
        return PJoin(self.source, pmer, other.source)

    def partition_reduce(self, f):
        """
        Provides medium-level functionality for partition reductions.  See 
        PMap.partition_reduce for more details.
        """
        reducer = StreamReducer(f)
        name = str(reducer)
        new_source, pmer = me.pmer._add_reducer([self.pm.source], 
                reducer, 
                name=name)

        return PMap(new_source, pmer)

class PJoin(PBase):
    """
    Performs different types of joins between two grouped datasets.
    """

    def __init__(self, source, pmer, right):
        super(PJoin, self).__init__(source, pmer)
        self.right = right

    def run(self, name=None, **kwargs):
        return self.reduce(lambda l, r: (list(l), list(r))).run(name, **kwargs)

    def reduce(self, aggregate, many=False):
        """
        Performs an inner join between two datasets.  The aggregate function
        will be provied two arguments: the left iterator and the right iterator.


            >>> left = Dampr.memory([("foo", 13), ("bar", 14)]).group_by(lambda x: x[0])
            >>> right = Dampr.memory([("bar", "baller"), ("baz", "bag")]).group_by(lambda x: x[0])
            >>> left.join(right).reduce(lambda lit, rit: (list(lit), list(rit))).read()
            [('bar', ([('bar', 14)], [('bar', 'baller')]))]


        If `many` is True, reduce will flatten the output into seperate records:

            >>> left.join(right).reduce(lambda lit, rit: list(lit) + list(rit), many=True).read()
            [('bar', ('bar', 14)), ('bar', ('bar', 'baller'))]
        """
        def _reduce(k, left, right):
            return aggregate(left, right)

        source, pmer = self.pmer._add_reducer([self.source, self.right], 
                KeyedInnerJoin(_reduce, many))
        return PMap(source, pmer)

    def left_reduce(self, aggregate):
        """
        Performs a left join on two datasets.  In the case where the right dataset
        is missing the join key, it will call the aggregate function with an empty
        iterator.

            >>> left = Dampr.memory([("foo", 13), ("bar", 14)]).group_by(lambda x: x[0])
            >>> right = Dampr.memory([("bar", "baller"), ("baz", "bag")]).group_by(lambda x: x[0])
            >>> left.join(right).left_reduce(lambda lit, rit: (list(lit), list(rit))).read()
            [('bar', ([('bar', 14)], [('bar', 'baller')])), ('foo', ([('foo', 13)], []))]
        """
        def _reduce(k, left, right):
            return aggregate(left, right)

        source, pmer = self.pmer._add_reducer([self.source, self.right], 
                KeyedLeftJoin(_reduce))
        return PMap(source, pmer)

    def _cross(self, crosser):
        def _cross(k1, v1, k2, v2):
            return k1, crosser(v1, v2)

        source, pmer = self.pmer._add_reducer([self.source, self.right],
                KeyedCrossJoin(_cross))

        return PMap(source, pmer).map(lambda x: x[1])

class Dampr(object):
    """
    Entrypoint into the Dampr processing functions.
    """
    def __init__(self, graph=None, runner=None):
        if graph is None:
            graph = Graph()

        self.graph = graph 
        if runner is None:
            runner = MTRunner

        self.runner = runner

    @classmethod
    def memory(cls, items, partitions=50):
        """
        Create an in-memory dataset from the provided items.  `partitions` define how
        many initial functions there will be.

            >>> Dampr.memory([1,2,3,4,5])
        """
        mi = MemoryInput(list(enumerate(items)), partitions)
        source, ng = Graph().add_input(mi)
        return PMap(source, Dampr(ng))

    @classmethod
    def read_input(cls, *datasets):
        """
        Reads from the provided datasets.  When provided multiple datasets, it treats
        each as a separate partition.

        `read_input` can also take Chunker, which lazily returns a set of Datasets to
        operate over.

            >>> from dampr.inputs import MemoryInput
            >>> Dampr.read_input(MemoryInput(enumerate(range(5,10)), 2))
        """
        if len(datasets) == 1 and isinstance(datasets, Chunker):
            datasets = datasets[0]
        elif len(datasets) > 1:
            datasets = CatDataset(datasets)
        else:
            datasets = datasets[0]

        source, ng = Graph().add_input(datasets)
        return PMap(source, Dampr(ng))

    @classmethod
    def text(cls, fname, chunk_size=16*1024**2):
        """
        Reads a file or directory of files into Dampr.  Each record is assumed to 
        be newline delimited.

        When fname is a directory, it will walk the directory, collecting all
        files within it as part of the collection.  `text` accepts globs.
        
        `chunk_size` describes how big each map portion will be.

        Returns a PMap object.

            >>> Dampr.text('/tmp', chunk_size=64*1024**2)
        """
        return cls.read_input(PathInput(fname, chunk_size))

    @classmethod
    def json(cls, *args, **kwargs):
        """
        Convenience function which reads newline-delimited json records.
        """
        return cls.text(*args, **kwargs).map(json.loads)

    @classmethod
    def from_dataset(cls, dataset):
        """
        Typically not used, this will read the raw outputs of a Dampr stage 
        as an input.
        """
        assert isinstance(dataset, Chunker)
        source, ng = Graph().add_input(dataset)
        return PMap(source, Dampr(ng))

    @classmethod
    def run(self, *pmers, **kwargs):
        """
        Runs a graph or set of graphs.

            >>> foo = Dampr.memory([1,2,3,4,5])
            >>> bar = Dampr.memory([6,7,8,9,10])
            >>> left, right = Dampr.run(foo, bar)
            >>> left.read()
            [1, 2, 3, 4, 5]
            >>> right.read()
            [6, 7, 8, 9, 10]
        """
        sources = []
        graph = None
        for i, pmer in enumerate(pmers):
            if isinstance(pmer, PMap):
                pmer = pmer.checkpoint()
            elif isinstance(pmer, PJoin):
                pmer = pmer.reduce(lambda l, r: (list(l), list(r)))
            
            if i == 0:
                graph = pmer.pmer.graph
            else:
                graph = pmer.pmer.graph.union(graph)

            sources.append(pmer.source)

        name = kwargs.pop('name', 'dampr/{}'.format(random.random()))
        ds = pmer.pmer.runner(name, graph, **kwargs).run(sources)
        return [ValueEmitter(d) for d in ds]

    def _add_mapper(self, *args, **kwargs): 
        output, ng = self.graph.add_mapper(*args, **kwargs)
        return output, Dampr(ng)

    def _add_reducer(self, *args, **kwargs): 
        output, ng = self.graph.add_reducer(*args, **kwargs)
        return output, Dampr(ng)

    def _add_sink(self, *args, **kwargs): 
        output, ng = self.graph.add_sink(*args, **kwargs)
        return output, Dampr(ng)

def fuse(aggs):
    if len(aggs) == 1:
        return aggs[0]

    s = aggs[1]
    for i in range(2, len(aggs)):
        s = ComposedStreamable(s, aggs[i])

    return ComposedMapper(aggs[0], s)

# This reinitializaes everytime
RANDOM = None
def get_rand():
    global RANDOM
    if RANDOM is None:
        RANDOM = random.Random(time.time())

    return RANDOM



import uuid
import os

from .dataset import *

class Splitter(object):
    def partition(self, key, n_partitions):
        return hash(key) % n_partitions

class Mapper(object):
    def map(self, *datasets):
        raise NotImplementedError()

class Streamable(object):
    def stream(self, kvs):
        raise NotImplementedError()

class Map(Mapper, Streamable):
    """
    Standard Mapper
    """
    def __init__(self, mapper):
        assert not isinstance(mapper, Mapper)
        self.mapper = mapper

    def map(self, *datasets):
        assert len(datasets) == 1
        return self.stream(datasets[0].read())

    def stream(self, kvs):
        for key, value in kvs:
            for k2, v2 in self.mapper(key, value):
                yield k2, v2

    def __unicode__(self):
        name = getattr(self.mapper, '__name__', str(type(self.mapper)))
        return u'Map[{}]'.format(name)

    __str__ = __unicode__
    __repr__ = __unicode__

class ComposedStreamable(Streamable):
    def __init__(self, left, right):
        assert isinstance(left, Streamable)
        assert isinstance(right, Streamable)
        self.left = left
        self.right = right

    def stream(self, kvs):
        return self.right.stream(self.left.stream(kvs))

class ComposedMapper(Mapper):
    def __init__(self, left, right):
        assert isinstance(left, Mapper)
        assert isinstance(right, Streamable)
        self.left = left
        self.right = right

    def map(self, *datasets):
        return self.right.stream(self.left.map(*datasets))

class BlockMapper(Mapper, Streamable):
    """
    Custom BlockMapper.  User's can specify how a Mapper instance can
    consume a stream of key-values, allowing for custom logic.
    """
    
    def start(self):
        """
        Sets up instance variables for when a Mapper begins consumption of 
        a Map Block.
        """
        pass

    def add(self, key, value):
        """
        Logic for how to handle new key-value pairs.  This function is required
        to return an iterator regardless of whether it yields data: this gives
        a more flexible definition.
        """
        raise NotImplementedError()

    def finish(self):
        """
        Mapping is finished.  In the case of aggregations, this should yield out
        all remaining key-values to consume.
        """
        return ()

    def map(self, *datasets):
        assert len(datasets) == 1
        return self.stream(datasets[0].read())

    def stream(self, kvs):
        self.start()
        for key, value in kvs:
            for out in self.add(key, value):
                yield out

        for out in self.finish():
            yield out

class StreamMapper(Mapper, Streamable):
    """
    Simple generator based block mapper.
    """

    def __init__(self, streamer_f):
        self.streamer_f = streamer_f
    
    def map(self, *datasets):
        assert len(datasets) == 1
        return self.stream(datasets[0].read())

    def stream(self, kvs):
        it = (v for _k, v in kvs)
        return self.streamer_f(it)

    def __unicode__(self):
        name = getattr(self.streamer_f, '__name__', str(type(self.streamer_f)))
        return u'StreamMapper[{}]'.format(self.streamer_f.__name__)

    __str__ = __unicode__
    __repr__ = __unicode__

class MapCrossJoin(Mapper):
    """
    Cross products two datasets.  If `cache` is True, will load up the
    right dataset into memory instead of streaming multiple times from
    disk.
    """
    def __init__(self, crosser, cache):
        self.crosser = crosser 
        self.cache = cache

    def map(self, *datasets):
        assert len(datasets) == 2
        left, right = [self.group_datasets(d) for d in datasets]

        # Cache the results
        if self.cache:
            cached = list(right.read())
            read_right = lambda: iter(cached)
        else:
            read_right = right.read

        for key, value in left.read():
            for key2, value2 in read_right():
                for k3, v3 in self.crosser(key, value, key2, value2):
                    yield k3, v3

    def group_datasets(self, dataset):
        if isinstance(dataset, Chunker):
            dataset = list(dataset.chunks())
        
        if len(dataset) > 1:
            dataset = CatDataset(dataset)
        elif len(dataset) == 1:
            dataset = dataset[0]
        else:
            dataset = EmptyDataset()
        
        return dataset

class Reducer(object):
    def reduce(self, *datasets):
        raise NotImplementedError()

    def yield_groups(self, dataset):
        return self.group_datasets(dataset).grouped_read()

    def group_datasets(self, dataset):
        if len(dataset) > 1:
            dataset = MergeDataset(dataset)
        elif len(dataset) == 1:
            dataset = dataset[0]
        else:
            dataset = EmptyDataset()
        
        return dataset

class Reduce(Reducer):
    """
    Standard reduce
    """
    def __init__(self, reducer):
        self.reducer = reducer

    def reduce(self, *datasets):
        assert len(datasets) == 1
        for k, vs in self.yield_groups(datasets[0]):
            yield k, self.reducer(k, vs)

class BlockReducer(Reducer):
    """
    Custom BlockReducer.  User's can specify how a Reducer instance can
    consume a stream of key-valueiters, allowing for custom logic.
    """
    def start(self):
        pass 

    def add(self, k, it):
        raise NotImplementedError()

    def finish(self):
        return ()

    def reduce(self, *datasets):
        assert len(datasets) == 1
        self.start()
        for k, vs in self.yield_groups(datasets[0]):
            for nk, nv in self.add(k, vs):
                yield nk, nv

        for nk, nv in self.finish():
            yield nk, nv

class StreamReducer(Reducer):
    """
    Custom BlockReducer.  User's can specify how a Reducer instance can
    consume a stream of key-valueiters, allowing for custom logic.
    """
    def __init__(self, stream_f):
        self.stream_f = stream_f

    def reduce(self, *datasets):
        assert len(datasets) == 1
        for nk, nv in self.stream_f(self.yield_groups(datasets[0])):
            yield nk, (nk, nv)

    def __unicode__(self):
        name = getattr(self.stream_f, '__name__', str(type(self.stream_f)))
        return u'StreamReducer[{}]'.format(self.stream_f.__name__)

    __str__ = __unicode__
    __repr__ = __unicode__


class KeyedReduce(Reduce):
    def reduce(self, *datasets):
        for k, v in super(KeyedReduce, self).reduce(*datasets):
            yield k, (k, v)

class InnerJoin(Reducer):
    def __init__(self, joiner_f, many=False):
        self.joiner_f = joiner_f
        self.many = many

    def reduce(self, *datasets):
        assert len(datasets) == 2
        g1 = self.yield_groups(datasets[0])
        g2 = self.yield_groups(datasets[1])
        left, right = next(g1, None), next(g2, None)
        while left is not None and right is not None:
            if left[0] < right[0]:
                left = next(g1, None)
            elif left[0] > right[0]:
                right = next(g2, None)
            else:
                k = left[0]
                it = self.joiner_f(k, left[1], right[1])
                if not self.many:
                    it = [it]

                for nv in it:
                    yield k, nv

                left, right = next(g1, None), next(g2, None)

class KeyedInnerJoin(InnerJoin):
    def reduce(self, *datasets):
        for k, v in super(KeyedInnerJoin, self).reduce(*datasets):
            yield k, (k, v)

class LeftJoin(Reducer):
    def __init__(self, joiner_f, default=lambda: iter([])):
        self.joiner_f = joiner_f
        self.default = default

    def reduce(self, *datasets):
        assert len(datasets) == 2
        g1 = self.yield_groups(datasets[0])
        g2 = self.yield_groups(datasets[1])
        left, right = next(g1, None), next(g2, None)
        while left is not None and right is not None:
            k = left[0]
            if left[0] < right[0]:
                yield k, self.joiner_f(k, left[1], self.default())
                left = next(g1, None)
            elif left[0] > right[0]:
                right = next(g2, None)
            else:
                yield k, self.joiner_f(k, left[1], right[1])
                left, right = next(g1, None), next(g2, None)

        # Finish off left
        while left is not None:
            k = left[0]
            yield k, self.joiner_f(k, left[1], self.default())
            left = next(g1, None)

class KeyedLeftJoin(LeftJoin):
    def reduce(self, *datasets):
        for k, v in super(KeyedLeftJoin, self).reduce(*datasets):
            yield k, (k, v)

class CrossJoin(Reduce):
    def __init__(self, joiner_f):
        self.joiner_f = joiner_f

    def reduce(self, *datasets):
        assert len(datasets) == 2
        for left in self.group_datasets(datasets[0]).read():
            for right in self.group_datasets(datasets[1]).read():
                yield self.joiner_f(left[0], left[1], right[0], right[1])

class KeyedCrossJoin(CrossJoin):
    def reduce(self, *datasets):
        for k, v in super(KeyedCrossJoin, self).reduce(*datasets):
            yield k, (k, v)

class OuterJoin(Reducer):
    def __init__(self, joiner_f, default=lambda: iter([])):
        self.joiner_f = joiner_f
        self.default = default

    def reduce(self, *datasets):
        assert len(datasets) == 2
        g1 = self.yield_groups(datasets[0])
        g2 = self.yield_groups(datasets[1])
        left, right = next(g1, None), next(g2, None)
        while left is not None and right is not None:
            if left[0] < right[0]:
                yield left[0], self.joiner_f(left[0], left[1], self.default())
                left = next(g1, None)
            elif left[0] > right[0]:
                yield right[0], self.joiner_f(right[0], iter([]), right[1])
                right = next(g2, None)
            else:
                yield k, self.joiner_f(k, left[1], right[1])
                left, right = next(g1, None), next(g2, None)

        # Finish off left
        while left is not None:
            yield left[0], self.joiner_f(left[0], left[1], self.default())
            left = next(g1, None)

        # Finish off right
        while right is not None:
            yield right[0], self.joiner_f(right[0], self.default(), right[1])
            right = next(g1, None)

class KeyedOuterJoin(OuterJoin):
    def reduce(self, *datasets):
        for k, v in super(KeyedOuterJoin, self).reduce(*datasets):
            yield k, (k, v)

class Combiner(object):
    """
    Interface for the combining ordered chunks from the Map stage
    """

    def combine(self, datasets):
        """
        Takes in a set of datasets and streams out key/values
        """
        raise NotImplementedError()

class NoopCombiner(Combiner):

    def combine(self, datasets):
       return MergeDataset(datasets)

class UnorderedCombiner(Combiner):
    def combine(self, datasets):
        return CatDataset(datasets)

class PartialReduceCombiner(Combiner):
    def __init__(self, reducer):
        self.reducer = reducer
    
    def _combine(self, datasets):
        for k, vs in MergeDataset(datasets).grouped_read():
            yield k, self.reducer.reducer(k, vs)

    def combine(self, datasets):
        return StreamDataset(self._combine(datasets))

class Shuffler(object):
    def __init__(self, n_partitions, splitter, writer_cls):
        self.n_partitions = n_partitions
        self.splitter = splitter
        self.writer_cls = writer_cls

    def shuffle(self, fs, datasets):
        """
        Needs to return a {partition_id: [datasets]}
        """
        raise NotImplementedError()

class DefaultShuffler(Shuffler):

    def shuffle(self, fs, datasets):
        partitions = []
        for i in range(self.n_partitions):
            writer = self.writer_cls(fs.get_substage('partition_{}'.format(i)))
            writer.start()
            partitions.append(writer)

        for k, v in MergeDataset(datasets).read():
            p_idx = self.splitter.partition(k, self.n_partitions)
            partitions[p_idx].add_record(k, v)

        splits = {}
        for i, writer in enumerate(partitions):
            splits[i] = writer.finished()[0]

        return splits

class FileSystem(object):
    def __init__(self, path):
        self.path = path

    def get_stage(self, name):
        return StageFileSystem(os.path.join(self.path, 'stage_{}'.format(name)))

class StageFileSystem(object):
    def __init__(self, path):
        self.path = path

    def get_worker(self, w_id):
        return WorkerFileSystem(os.path.join(self.path, 'worker_{}'.format(w_id)))

class WorkingFileSystem(object):
    def __init__(self, path):
        self.path = path 

    def get_file(self, name=None):
        if name is None:
            name = str(uuid.uuid4())

        if not os.path.isdir(self.path):
            os.makedirs(self.path)

        new_file = os.path.join(self.path, name)
        return new_file

class WorkerFileSystem(WorkingFileSystem):
            
    def get_substage(self, s):
        return SubStageFileSystem(os.path.join(self.path, 'sub_{}'.format(s)))

class SubStageFileSystem(WorkingFileSystem):
    pass
    



#!/usr/bin/env python
import os

from setuptools import setup, find_packages
from distutils.core import setup

BASE = os.path.dirname(os.path.abspath(__file__))

with open(os.path.join(BASE, "README.md")) as f:
    long_description = f.read()

setup(
    name='dampr',
    version="0.1.6",
    description="Data Processing implementation in Python",
    long_description=long_description,
    url="https://www.github.com/Refefer/Dampr",
    packages=find_packages(BASE),
    test_suite="tests",
    author='Andrew Stanton',
    author_email="refefer@gmail.com",
    classifiers=[
       "Development Status :: 3 - Alpha",
       "License :: OSI Approved :: Apache Software License",
       "Programming Language :: Python :: 2",
       "Programming Language :: Python :: 3",
       "Operating System :: OS Independent"
      ]
    )
